import asyncio as _asyncio
import logging as _logging
from abc import ABCMeta as _ABCMeta
from typing import Any, List

import aioxmpp as _aioxmpp
import spade as _spade
from aioxmpp import JID

_logger = _logging.getLogger(__name__)


class _XMPPAgent(_spade.agent.Agent):
    """Agent that integrates XMPP functionalities.

    Attributes:
        jid: XMPP identifier.
        verify_security: If true verifies the SSL certificates.
    """

    def __init__(self, jid: JID, verify_security: bool = False):
        _logging.getLogger(jid.localpart).setLevel(_logging.ERROR)
        self.groups = dict()
        self.muc_client = None
        pw = str(jid.bare())
        jid = str(jid)
        super().__init__(jid, pw, verify_security)

    async def _hook_plugin_after_connection(self):
        """Executed after SPADE Agent's connection.

        This method adds the MUC service to the Agent XMPP Client and
        adds a message dispatcher for the group(MUC) messages.
        """
        self.presence.approve_all = True

        self.muc_client: _aioxmpp.MUCClient = self.client.summon(_aioxmpp.MUCClient)
        self.disco: _aioxmpp.DiscoClient = self.client.summon(_aioxmpp.DiscoClient)

        self.message_dispatcher.register_callback(
            _aioxmpp.MessageType.GROUPCHAT,
            None,
            self._message_received,
        )
        self.message_dispatcher.register_callback(
            _aioxmpp.MessageType.NORMAL,
            None,
            self._message_received,
        )

    def _on_muc_failure_handler(self, exc):
        """Handles MUC failed connections."""

        _logger.critical("Failed to enter MUC room")
        raise exc

    async def join_group(self, jid: str):
        """Joins the agent in a group.

        Args:
            jid: Group's XMPP identifier.

        Raises:
            Exception if the group JID is invalid.
        """
        room, fut = self.muc_client.join(_aioxmpp.JID.fromstr(jid), self.name)
        room.on_failure.connect(self._on_muc_failure_handler)
        await _asyncio.wait([fut], timeout=3)
        if fut.done():
            if jid not in self.groups:
                self.groups[jid] = room
                _logger.info("joined group: " + jid)
        else:
            raise Exception("invalid group JID: " + str(jid))

    async def leave_group(self, jid: str):
        """Leaves a group.

        Args:
            jid: Group's XMPP identifier.
        """
        room = self.groups.pop(jid, None)
        if room:
            _logger.info("leaving group: " + jid)
            await room.leave()

    async def list_groups(self, node_jid: str):
        """Retrieves the list of the existing groups.

        Args:
            node_jid: JID of the domain responsable for the MUC functionality.

        Returns:
            A list of XMPP groups.
        """
        info = await self.disco.query_items(
            _aioxmpp.JID.fromstr(node_jid), require_fresh=True
        )
        return info.items

    async def group_members(self, jid: str) -> List:
        """Extracts list of group members from a group.

        Args:
            jid: Group's XMPP identifier.

        Returns:
            A copy of the list of occupants. The local user is always the first item in the list.
        """
        if jid in self.groups:
            return self.groups[jid].members
        else:
            await self.join_group(jid)
            members = self.groups[jid].members
            await self.leave_group(jid)
            return members


class Agent(_XMPPAgent):
    """PEAK's base agent.

    Attributes:
        jid: XMPP identifier.
        properties: Properties to be injected in the agent.
        verify_security: If True, it verifies the SSL certificates.
    """

    def __init__(self, jid: JID, properties: Any = None, verify_security: bool = False):
        """Inits Agent and fills it with properties."""
        super().__init__(jid, verify_security=verify_security)
        if properties:
            self.properties = properties
            for key in properties:
                setattr(self, key, properties[key])

    def iterate_properties(self):
        """Iterates one index over the properties."""
        if hasattr(self, "properties"):
            for key in self.properties:
                attr = getattr(self, key)
                if attr:
                    getattr(self, key).next()


class _Behaviour:
    """Adds functinality to the SPADE base behaviours."""

    agent: Agent

    async def send_to_group(self, msg: _spade.message.Message):
        """Sends a message to a group.

        If the agent is not in the group, the agent enters the room first,
        sends the message and then leaves the group.

        Args:
            msg: The XMPP message.
        """
        raw_msg = msg.prepare()
        _logger.debug(f"sending message '{msg.body}' to {msg.to}")
        try:
            await self.agent.groups[str(msg.to)].send_message(raw_msg)
        except:
            _logger.debug(f"agent not member of {msg.to}")
            room, future = self.agent.muc_client.join(msg.to, self.agent.name)
            await future
            await room.send_message(raw_msg)
            await room.leave()
            _logger.debug(f"leaving {msg.to}")
    
    async def execute(self, behaviour):
        """Executes and awaits synchronozly for a behaviour.

        It is used to chain behviour that are dependent on each other.

        Args:
            behaviour: SPADE's behaviour.
        """
        self.agent.add_behaviour(behaviour)
        await behaviour.join()

class OneShotBehaviour(
    _spade.behaviour.OneShotBehaviour, _Behaviour, metaclass=_ABCMeta
):
    pass


class PeriodicBehaviour(
    _spade.behaviour.PeriodicBehaviour, _Behaviour, metaclass=_ABCMeta
):
    pass


class CyclicBehaviour(_spade.behaviour.CyclicBehaviour, _Behaviour, metaclass=_ABCMeta):
    pass


class FSMBehaviour(_spade.behaviour.FSMBehaviour, _Behaviour, metaclass=_ABCMeta):
    pass



import logging
from abc import ABCMeta

from pyModbusTCP.client import ModbusClient
from pyModbusTCP.utils import get_2comp

log = logging.getLogger("peak.mas.drivers")


class Driver(metaclass=ABCMeta):
    """Base interface for the Drivers."""

    pass


class DriverModBusTCP(Driver):
    """Protocol ModBus/TCP.

    Attributes:
        client: handles the ModBus/TCP comunication.
    """

    def __init__(self, host: str, unit_id: int = None):
        """Inits Driver ModBus/TCP.

        Args:
            host: Hostname or IPv4/IPv6 address of the device.
            unit_id: Unit ID.
        """
        self.client = ModbusClient(host=host, unit_id=unit_id, auto_open=True)

    def read(self, register: int, unit_id: int = None) -> int:
        """Reads a given register of the device.

        Args:
            register: Register address (0 to 65535).
            unit_id: Unit ID.

        Returns:
            An integer that was read at the given register.
            Returns None if the value could not read from the
            target.
        """
        if unit_id is not None:
            self.client.unit_id(unit_id)
        data = self.client.read_input_registers(reg_addr=register)
        if not data:
            log.warning(
                "%s cannot read from target %s.",
                self.__class__,
                self.client.host(),
                exc_info=self.client.last_except_txt(),
            )
            return None
        else:
            return get_2comp(data[0])

    def write(self, register: int, value: int, unit_id: int = None):
        """Writes a value in a given register.

        Args:
            register: Register address (0 to 65535).
            value: Value to be written.
            unit_id: Unit ID.
        """
        if unit_id is not None:
            self.client.unit_id(unit_id)
        if not self.client.write_single_register(register, get_2comp(value)):
            log.warning(
                "%s cannot write to target %s.",
                self.__class__,
                self.client.host(),
                exc_info=self.client.last_except_txt(),
            )



import logging
from json import dumps as json_dumps
from typing import Callable

from peak import DF, Message, SyncAgent, Template
from peak.core import CyclicBehaviour, OneShotBehaviour, PeriodicBehaviour
from peak.properties import Property


class JoinGroup(OneShotBehaviour):
    """Joins a group using a JID."""

    def __init__(self, path: str, domain: str, tags: list = []):
        super().__init__()
        self.path = path
        self.domain = domain
        self.tags = tags

    async def run(self):
        msg = Message()
        msg.to = DF.name(self.agent.jid.domain)
        msg.set_metadata("resource", "treehierarchy")
        msg.set_metadata("path", self.path)
        msg.set_metadata("domain", self.domain)
        msg.set_metadata("tags", str(self.tags))
        await self.send(msg)
        nodes = self.path.split("/")
        for node in nodes[:-1]:
            await self.agent.join_group(node + "_down@" + self.domain)
        await self.agent.join_group(nodes[-1] + "@" + self.domain)
        await self.agent.join_group(nodes[-1] + "_down@" + self.domain)


class LeaveGroup(OneShotBehaviour):
    """Leaves a group."""

    def __init__(self, path: str, domain: str):
        super().__init__()
        self.path = path
        self.domain = domain

    async def run(self):
        msg = Message()
        msg.to = DF.name(self.agent.jid.domain)
        msg.set_metadata("resource", "treehierarchy")
        msg.set_metadata("path", self.path)
        msg.set_metadata("domain", self.domain)
        msg.set_metadata("leave", "true")
        await self.send(msg)
        nodes = self.path.split("/")
        for node in nodes[:-1]:
            await self.agent.leave_group(node + "_down@" + self.domain)
        await self.agent.leave_group(nodes[-1] + "@" + self.domain)
        await self.agent.leave_group(nodes[-1] + "_down@" + self.domain)


class SearchGroup(OneShotBehaviour):
    """Searches for a group."""

    def __init__(
        self, tags: list[str], callback: Callable[[list[str]], None], *args, **kargs
    ):
        self.tags = tags
        self.callback = callback
        self.args = args
        self.kargs = kargs

    async def run(self):
        msg = Message()
        msg.to = DF.name(self.agent.jid.domain)
        msg.set_metadata("resource", "searchgroup")
        msg.set_metadata("tags", str(self.tags))
        await self.send(msg)
        res = None
        while not res:
            res = await self.receive(60)
            if not res:
                raise Exception("DF did not respond")
            groups = res.get_metadata("groups")
            logging.getLogger(self.__class__.__name__).info(self.tags, groups)
            self.callback(self.tags, groups, *self.args, **self.kargs)


class ExportData(OneShotBehaviour):
    """Exports data to a file and optionaly to the DF.

    The data exports works diferently when using the Synchronizer and
    when its not used. When the Synchronizer its being used the data
    is exported at the same rate as the Synchronizers clock. If its not
    used the user must define a interval in which the data is exported.
    
    Warning: the file will be overwriten."""

    def __init__(
        self,
        file_name: str,
        properties: list,
        interval: int = None,
        to_graph: bool = False,
        graph_name: str = "",
        graph_options: str = "",
    ):
        super().__init__()
        self.interval = interval
        self.file_name = file_name
        self.properties = properties
        self.to_graph = to_graph
        if graph_name == "" and self.to_graph:
            raise Exception("when graph is set to true, a name for it must be defined")
        self.graph_name = graph_name
        self.graph_options = graph_options

    async def run(self) -> None:
        logger = logging.getLogger(self.__class__.__name__)
        if isinstance(self.agent, SyncAgent):
            logger.debug("synchronizer detected")
            self.agent.add_behaviour(
                _ExportDataSync(
                    self.file_name,
                    self.properties,
                    self.to_graph,
                    self.graph_name,
                    self.graph_options,
                )
            )
        else:
            logger.debug("synchronizer not detected")
            self.agent.add_behaviour(
                _ExportData(
                    self.file_name,
                    self.properties,
                    self.interval,
                    self.to_graph,
                    self.graph_name,
                    self.graph_options,
                )
            )


class _ExportDataSync(CyclicBehaviour):
    """Exports the data at the same rate as the synchronization."""

    def __init__(
        self,
        file_name: str,
        properties: list,
        to_graph: bool,
        graph_name: str,
        graph_options: str,
    ):
        super().__init__()
        self.file_name = file_name
        self.graph_name = graph_name
        self.properties = properties
        self.to_graph = to_graph
        self.graph_options = graph_options

    async def on_start(self) -> None:
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.info("behaviour initiated")
        template_step = Template()
        template_step.set_metadata("sync", "step")
        template_stop = Template()
        template_stop.set_metadata("sync", "stop")
        template = template_step | template_stop
        self.set_template(template)

        self.file_data = {}
        for property in self.properties:
            self.file_data[property] = []
        if self.to_graph:
            msg = Message(to=DF.name(self.agent.jid.domain))
            msg.body = "Create graph " + self.graph_name
            msg.metadata = {
                "resource": "graph",
                "action": "create",
                "graph_name": self.graph_name,
                "graph_options": self.graph_options,
                "properties": json_dumps(self.properties),
            }
            await self.send(msg)
            self.logger.info("creating graph " + +self.graph_name)

    async def run(self) -> None:
        msg = await self.receive(60)
        if msg:
            if msg.get_metadata("sync") == "step":
                current_data = dict()
                for property in self.file_data:
                    attribute = getattr(self.agent, property)
                    if type(attribute) is Property:
                        self.file_data[property].append(attribute.current_value)
                        current_data[property] = attribute.current_value
                    else:
                        self.file_data[property].append(attribute)
                        current_data[property] = attribute
                if self.to_graph:
                    msg = Message(to=DF.name(self.agent.jid.domain))
                    msg.body = "Update graph " + self.file_name
                    msg.metadata = {
                        "resource": "graph",
                        "action": "update",
                        "graph_name": self.file_name,
                        "data": json_dumps(current_data),
                    }
                    await self.send(msg)
                    self.logger.info("updating graph " + +self.graph_name)
            if msg.get_metadata("sync") == "stop":
                with open(self.file_name, "w") as f:
                    f.write(json_dumps(self.data))


class _ExportData(PeriodicBehaviour):
    """Exports the data at a rate defined by the user."""

    def __init__(
        self,
        file_name: str,
        properties: list,
        interval: float,
        to_graph: bool,
        graph_name: str,
        graph_options: str,
    ):
        super().__init__(interval, None)
        self.file_name = file_name
        self.graph_name = graph_name
        self.properties = properties
        self.to_graph = to_graph
        self.graph_options = graph_options

    async def on_start(self) -> None:
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.info("exporting " + str(self.properties))
        self.file_data = {}
        for property in self.properties:
            self.file_data[property] = []
        if self.to_graph:
            msg = Message(to=DF.name(self.agent.jid.domain))
            msg.body = "Create graph " + self.graph_name
            msg.metadata = {
                "resource": "graph",
                "action": "create",
                "graph_name": self.graph_name,
                "graph_options": self.graph_options,
                "properties": json_dumps(self.properties),
            }
            await self.send(msg)
            self.logger.info("creating graph " + self.graph_name)

    async def run(self) -> None:
        current_data = dict()
        for property in self.file_data:
            attribute = getattr(self.agent, property)
            if type(attribute) is Property:
                self.file_data[property].append(attribute.current_value)
                current_data[property] = attribute.current_value
            else:
                self.file_data[property].append(attribute)
                current_data[property] = attribute
            self.logger.debug(
                'exporting property "' + property + '": ' + str(current_data[property])
            )
        if self.to_graph:
            msg = Message(to=DF.name(self.agent.jid.domain))
            msg.body = "Update graph " + self.graph_name
            msg.metadata = {
                "resource": "graph",
                "action": "update",
                "graph_name": self.graph_name,
                "data": json_dumps(current_data),
            }
            await self.send(msg)
            self.logger.debug("updating graph " + self.graph_name)

    async def on_end(self) -> None:
        self.logger.info('exporting to file "' + self.file_name + '"')
        with open(self.file_name, "w") as f:
            f.write(json_dumps(self.file_data))



import logging
from argparse import ArgumentParser
from pathlib import Path
import sys

from peak import JID
from peak import __name__ as peak_name
from peak.cli import df, mas


logger = logging.getLogger()

def main(args=None):
    parser = ArgumentParser(prog=peak_name)
    parser.add_argument("-version", action="version", version="PEAK 1.0")
    parser.add_argument(
        "-v",
        action="store_true",
        help="Verbose. Turns on the debug info.",
    )
    subparsers = parser.add_subparsers(required=True)

    # parser for "df" command
    df_parser = subparsers.add_parser(
        name="df", help="Execute Directory Facilitator agent."
    )
    df_parser.add_argument(
        "-domain",
        type=str,
        default="localhost",
        help="XMPP domain to which the DF must register and login. Default is localhost.",
    )
    df_parser.add_argument(
        "--verify_security", action="store_true", help="Verifies the SLL certificates."
    )
    df_parser.add_argument(
        "-log_level",
        type=lambda x: logging.getLevelName(str.upper(x)),
        default=logging.getLevelName("INFO"),
        help="Selects the logging level of the DF. Default is INFO.",
    )
    df_parser.add_argument(
        "-port", type=str, default="10000", help="Port to be opened for the REST API. Default is 10000"
    )
    df_parser.set_defaults(func=df.exec)

    # parser for the "run" command
    run_parser = subparsers.add_parser(
        name="run",
        help="Execute PEAK's agents.",
    )
    run_parser.add_argument(
        "file",
        type=Path,
        help="This file must be a python file. The Python file must contain a class of a single agent to be executed. The name of the class must be the same of the name of the file. ",
    )
    run_parser.add_argument(
        "-jid", type=JID.fromstr, help="JID of the agent to be executed.", required=True
    )
    run_parser.add_argument(
        "-properties",
        type=Path,
        help="Python file where the agent properties are located.",
    )
    run_parser.add_argument(
        "-clone",
        type=int,
        default=1,
        help="The number of clones of the agent. The name of each agent will be the same as the JID but with the number of the corresponding clone to it (e.g. john_0@localhost, john_1@localhost). The sequence starts in zero.",
    )
    run_parser.add_argument(
        "-log_level",
        type=lambda x: logging.getLevelName(str.upper(x)),
        default=logging.getLevelName("INFO"),
        help="Selects the logging level of the agent.",
    )
    run_parser.add_argument(
        "--verify_security", action="store_true", help="Verifies the SLL certificates."
    )
    run_parser.set_defaults(func=mas.agent_exec)

    # parser for the "run" command
    start_parser = subparsers.add_parser(
        name="start",
        help="Executes multiple agents using a YAML configuration file.",
    )
    start_parser.add_argument(
        "file",
        type=Path,
        help="Path to the YAML configuration file.",
    )
    start_parser.set_defaults(func=mas.multi_agent_exec)
    
    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args(args)
    if args.v:
        logger.setLevel("DEBUG")
    args.func(**vars(args))


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        logger.exception(e)
    except KeyboardInterrupt:
        pass



import asyncio as _asyncio
import logging as _logging
from abc import ABCMeta as _ABCMeta
from abc import abstractmethod
from datetime import datetime, timedelta
from json import loads as json_loads

import aiohttp_cors
from aioxmpp import JID

from peak import Agent, CyclicBehaviour, Message, PeriodicBehaviour, Template
from peak.properties import Properties

_logger = _logging.getLogger(__name__)


class SyncAgent(Agent, metaclass=_ABCMeta):
    """Agent that is synchronized by the Synchronizer.

    Every agent that needs to be synchronized needs
    to extend this class.

    Attributes:
        period: An integer representing the number of the period inside the simulation.
        time: A datetime of the current moment in the simmulation.
    """

    class _StepBehaviour(CyclicBehaviour):
        """Listens for the Synchronizer's clock's ticks."""

        async def on_start(self):
            _logger.info("Waiting for simulation to start...")

        async def run(self):
            msg = await self.receive(10)
            if msg:
                if msg.get_metadata("sync") == "step":
                    self.agent.period = int(msg.get_metadata("period"))
                    self.agent.time = datetime.strptime(
                        msg.get_metadata("time"), "%Y-%m-%d %H:%M:%S"
                    )
                    if self.agent.period != 0:
                        self.agent.iterate_properties()
                    await self.agent.step()
                    _logger.info(
                        "Period "
                        + str(self.agent.period)
                        + " ("
                        + str(self.agent.time)
                        + ")"
                    )
                if msg.get_metadata("sync") == "stop":
                    _logger.info("Simulation ended")
                    self.kill()

        async def on_end(self):
            await self.agent.stop()

    def __init__(
        self, jid: JID, properties: Properties = None, verify_security: bool = False
    ):
        """Agent that listens to the Synchronizer.

        Args:
            jid: XMPP identifier of the agent.
            properties: Atributes of the agent to be injected.
            verify_security: If True verifies SSL certificates.
        """
        super().__init__(jid, properties, verify_security)
        self.period = 0
        self.time = None
        template_step = Template()
        template_step.set_metadata("sync", "step")
        template_stop = Template()
        template_stop.set_metadata("sync", "stop")
        template = template_step | template_stop
        self.add_behaviour(self._StepBehaviour(), template)

    @abstractmethod
    async def step(self):
        """Executed at each tick of the Synchronizer clock.

        To be implemented by the user."""
        raise NotImplementedError()


class Synchronizer(Agent):
    """Agent that synchronizes the other agents.

    The Synchronizer creates a group of agents, awaits for
    the agents to join the group and starts the clock of the
    simulation.
    """

    class _PeriodicClock(PeriodicBehaviour):
        """Handles the clock of the simulation.

        This clock tracks the number of the current period
        throughout the simulation.
        """

        def __init__(
            self,
            group_jid: str,
            n_agents: int,
            periods: int,
            time_per_period: float,
            start_at=None,
        ):
            super().__init__(time_per_period, start_at=start_at)
            self.group_jid = group_jid
            self.n_agents = n_agents
            self.periods = periods

        async def on_start(self):
            while not len(self.agent.group_members(self.group_jid)) >= self.n_agents:
                await _asyncio.sleep(1)
            self.current_period = 0
            _logger.info("Starting simulation...")

        async def run(self):
            msg = Message()
            msg.to = self.group_jid
            if self.current_period >= self.periods:
                msg.set_metadata("sync", "stop")
                self.kill()
            else:
                _logger.info("Period " + str(self.current_period))
                msg.body = "Period " + str(self.current_period)
                msg.set_metadata("sync", "step")
                msg.set_metadata("period", str(self.current_period))
            await self.send_to_group(msg)
            self.current_period += 1

        async def on_end(self):
            _logger.info("Ending simulation...")
            await self.agent.stop()

    class _DateTimeClock(PeriodicBehaviour):
        """Handles the clock of the simulation.

        This clock tracks the current date and time of the
        simulation throughout its execution.
        """

        def __init__(
            self,
            jid,
            n_agents: int,
            initial_time: datetime,
            end_time: datetime,
            period_time_simulated: timedelta,
            period_time_real: float,
            start_at: datetime = None,
        ):
            super().__init__(period_time_real, start_at=start_at)
            self.group_jid = jid
            self.n_agents = n_agents
            self.time = initial_time
            self.end_time = end_time
            self.period_time = period_time_simulated

        async def on_start(self):
            _logger.info("Waiting for all agents to enter the group...")
            while (
                not len(await self.agent.group_members(self.group_jid)) >= self.n_agents
            ):
                await _asyncio.sleep(1)
            self.current_period = 0
            _logger.info("Starting simulation...")

        async def run(self):
            msg = Message()
            msg.to = self.group_jid
            if self.time >= self.end_time:
                msg.set_metadata("sync", "stop")
                self.kill()
            else:
                _logger.info(
                    "Period " + str(self.current_period) + " (" + str(self.time) + ")"
                )
                msg.body = (
                    "Period " + str(self.current_period) + " (" + str(self.time) + ")"
                )
                msg.set_metadata("sync", "step")
                msg.set_metadata("period", str(self.current_period))
                msg.set_metadata(
                    "time", datetime.strftime(self.time, "%Y-%m-%d %H:%M:%S")
                )
            await self.send_to_group(msg)
            self.current_period += 1
            self.time += self.period_time

        async def on_end(self):
            _logger.info("Ending simulation...")
            await self.agent.stop()

    async def sync_group_period(
        self, group_jid: str, n_agents: int, time_per_period: float, periods: int
    ):
        """Synchronizes a group of agents.

        The clock is based on the number of the current period.

        Args:
            group_jid: Identifier of the XMPP group to be synchronized.
            n_agents: Number of agents to be synchronized. Synchronizer
                awaits for this number of agents to join the group before
                it starts the simulation.
            time_per_period: time in seconds between each period.
            periods: number of periods to simulate.
        """
        await self.join_group(group_jid)
        self.add_behaviour(
            self._PeriodicClock(group_jid, n_agents, periods, time_per_period)
        )

    async def sync_group_time(
        self,
        group_jid: str,
        n_agents: int,
        initial_time: datetime,
        end_time: datetime,
        internal_period_time: timedelta,
        external_period_time: float,
        start_at: datetime = None,
    ):
        """Synchronizes a group of agents.

        The clock is based on the date and time inside of the simulation.

        Args:
            group_jid: Identifier of the XMPP group to be synchronized.
            n_agents: Number of agents to be synchronized. Synchronizer
                awaits for this number of agents to join the group before
                it starts the simulation.
            initial_time: defines the internal date and time at the start of the
                simulation.
            end_time: defines the internal date and time at which the simulation ends.
            internal_period_time: time between each period inside the simulation.
            period_time_real: time between each tick of the clock.
            start_at: schedules the simulation start at a given time. If None the
                simulation starts right away.
        """
        await self.join_group(group_jid)
        self.add_behaviour(
            self._DateTimeClock(
                group_jid,
                n_agents,
                initial_time,
                end_time,
                internal_period_time,
                external_period_time,
                start_at,
            )
        )


class DF(Agent):
    """Directory Facilitator.

    This agent makes available the data that it gathers from multi-agent systems
    publicly through a REST API. It also provides a Yellow Page Service to the
    agents.
    """

    class _GroupHierarchy(CyclicBehaviour):
        """Manages the group structure of all the multi-agent systems."""

        async def on_start(self):
            self.logger = _logging.getLogger(self.__class__.__name__)
            self.logger.debug("starting behaviour")
            template = Template()
            template.set_metadata("resource", "treehierarchy")
            self.set_template(template)
            self.agent.grouphierarchy_data = {
                "nodes": set(),
                "links": set(),
                "categories": set(),
                "node_members": {},
                "tags": {},
            }

        async def run(self):
            msg = await self.receive(60)
            if msg:
                self.logger.debug("message received")
                path = msg.get_metadata("path")
                domain = msg.get_metadata("domain")
                if meta_tags := msg.get_metadata("tags"):
                    tags = json_loads(meta_tags)
                nodes = path.split("/")
                self.logger.debug("nodes: " + str(nodes))
                level = "level"

                if (
                    msg.get_metadata("leave")
                    and msg.sender
                    in self.agent.grouphierarchy_data["node_members"][nodes[-1]]
                ):
                    self.logger.debug(str(msg.sender) + " leaving " + path)
                    self.agent.grouphierarchy_data["node_members"][nodes[-1]].remove(
                        msg.sender
                    )
                    nodes = nodes[::-1]
                    # remove empty nodes and links
                    for i, node in enumerate(nodes):
                        if len(
                            self.agent.grouphierarchy_data["node_members"][node]
                        ) == 0 and not any(
                            node == source
                            for source, _ in self.agent.grouphierarchy_data["links"]
                        ):
                            self.agent.grouphierarchy_data["node_members"].pop(node)
                            self.agent.grouphierarchy_data["nodes"].remove(
                                (node, level + str(len(nodes) - 1 - i), domain)
                            )
                            if i + 1 < len(nodes):
                                self.agent.grouphierarchy_data["links"].remove(
                                    (nodes[i + 1], node)
                                )
                    existing_categories = set()

                    # remove categories if empty
                    for _, level, _ in self.agent.grouphierarchy_data["nodes"]:
                        existing_categories.add(level)
                    difference = self.agent.grouphierarchy_data[
                        "categories"
                    ].difference(existing_categories)
                    if any(difference):
                        self.agent.grouphierarchy_data["categories"] -= difference

                else:
                    self.logger.debug(str(msg.sender) + " entering " + path)
                    last = None
                    for i, node in enumerate(nodes):
                        if node not in self.agent.grouphierarchy_data["node_members"]:
                            self.agent.grouphierarchy_data["node_members"][node] = []
                        self.agent.grouphierarchy_data["nodes"].add(
                            (node, level + str(i), domain)
                        )
                        self.agent.grouphierarchy_data["categories"].add(level + str(i))
                        if last != None:
                            self.agent.grouphierarchy_data["links"].add(
                                (
                                    last,
                                    node,
                                    max(
                                        len(
                                            self.agent.grouphierarchy_data[
                                                "node_members"
                                            ][last]
                                        )
                                        + 1,
                                        len(
                                            self.agent.grouphierarchy_data[
                                                "node_members"
                                            ][node]
                                        ),
                                    ),
                                )
                            )
                        last = node
                    self.agent.grouphierarchy_data["node_members"][last].append(
                        msg.sender
                    )
                    for tag in tags:
                        if tag not in self.agent.grouphierarchy_data["tags"]:
                            self.agent.grouphierarchy_data["tags"][tag] = set()
                        self.agent.grouphierarchy_data["tags"][tag].add(last)

    class _SearchGroup(CyclicBehaviour):
        """Handles all the requests to search for groups."""

        async def on_start(self) -> None:
            self.logger = _logging.getLogger(self.__class__.__name__)
            self.logger.debug("starting behaviour")
            template = Template()
            template.set_metadata("resource", "searchgroup")
            self.set_template(template)

        async def run(self) -> None:
            msg = await self.receive(60)
            if msg:
                tags = json_loads(msg.get_metadata("tags"))
                groups = self.agent.grouphierarchy_data["tags"][tags[0]]
                for tag in tags[1:]:
                    groups.intersection(self.agent.grouphierarchy_data["tags"][tag])
                res = msg.make_reply()
                res.set_metadata("groups", groups)
                await self.send(res)

    class _CreateGraph(CyclicBehaviour):
        """Handles the requests to create graphs"""

        async def on_start(self):
            self.logger = _logging.getLogger(self.__class__.__name__)
            self.logger.debug("starting behaviour")
            template = Template()
            template.set_metadata("resource", "graph")
            template.set_metadata("action", "create")
            self.set_template(template)

        async def run(self) -> None:
            msg = await self.receive(60)
            if msg:
                self.logger.debug(msg.body)
                graph_name = msg.get_metadata("graph_name")
                graph_options = msg.get_metadata("graph_options")
                properties = json_loads(msg.get_metadata("properties"))
                self.agent.dataanalysis_data[graph_name] = {
                    "graph_options": graph_options,
                    "data": {},
                }
                for property in properties:
                    self.agent.dataanalysis_data[graph_name]["data"][property] = []

    class _UpdateGraph(CyclicBehaviour):
        """Handles the requests to update the data of a given graph."""

        async def on_start(self) -> None:
            self.logger = _logging.getLogger(self.__class__.__name__)
            self.logger.debug("starting behaviour")
            template = Template()
            template.set_metadata("resource", "graph")
            template.set_metadata("action", "update")
            self.set_template(template)

        async def run(self) -> None:
            msg = await self.receive(60)
            if msg:
                self.logger.debug(msg.body)
                graph_name = msg.get_metadata("graph_name")
                data = json_loads(msg.get_metadata("data"))
                for property in data:
                    self.agent.dataanalysis_data[graph_name]["data"][property].append(
                        data[property]
                    )
                    self.logger.debug(
                        'updating property "' + property + '" : ' + str(data[property])
                    )

    def __init__(self, domain, verify_security, port):
        super().__init__(
            JID.fromstr("df@" + domain + "/admin"), verify_security=verify_security
        )
        self.port = port

    @classmethod
    def name(cls, domain: str) -> str:
        """Retrieves the JID of the DF based on the agent domain.

        Args:
            domain: server's name

        Returns:
            The string of the complete JID of the DF.
        """
        return "df@" + domain + "/admin"

    async def setup(self):
        self.grouphierarchy_data = dict()
        self.dataanalysis_data = dict()
        self.group_tags = dict()

        self.add_behaviour(self._GroupHierarchy())
        self.add_behaviour(self._SearchGroup())
        self.add_behaviour(self._CreateGraph())
        self.add_behaviour(self._UpdateGraph())

        # Creates routes.
        self.web.add_get("/groups", self.groups, template=None)
        self.web.add_get("/groups/refresh", self.refresh_groups, template=None)
        self.web.add_get("/dataanalysis", self.dataanalysis, template=None)

        # Configure default CORS settings.
        cors = aiohttp_cors.setup(
            self.web.app,
            defaults={
                "*": aiohttp_cors.ResourceOptions(
                    allow_credentials=True,
                    expose_headers="*",
                    allow_headers="*",
                )
            },
        )

        # Configure CORS on all routes.
        for route in list(self.web.app.router.routes()):
            cors.add(route)

        # Start web API
        self.web.start(port=self.port)
        _logger.info("REST API running on port " + self.port)

    async def groups(self, request):
        graph = {
            "nodes": list(self.grouphierarchy_data["nodes"]),
            "links": list(self.grouphierarchy_data["links"]),
            "categories": list(self.grouphierarchy_data["categories"]),
            "node_members": self.grouphierarchy_data["node_members"],
        }
        return graph

    async def refresh_groups(self, request):
        pass

    async def dataanalysis(self, request):
        return self.dataanalysis_data



"""PEAK - Python-based framework for hEterogeneous Agent Communities.

PEAK is a framework for developing communities of heterogeneous agents. 
This communities are multi-agent systems that can coexist and exchange 
resources and information with each other easly. 

isort: skip_file
"""

import logging as _logging

from spade.message import Message
from spade.template import Template

from peak.core import *
from peak.agents import *
from peak.behaviours import *
from peak.properties import *

__author__ = "Bruno Ribeiro"
__email__ = "brgri@isep.ipp.pt"
__version__ = "1.0.2"

# filtering noisy loggers
_logging.getLogger("aiosasl").setLevel(_logging.ERROR)
_logging.getLogger("spade").setLevel(_logging.ERROR)
_logging.getLogger("muc").setLevel(_logging.ERROR)
_logging.getLogger("StanzaStream").setLevel(_logging.ERROR)
_logging.getLogger("XMLStream").setLevel(_logging.ERROR)
_logging.getLogger("aioopenssl").setLevel(_logging.ERROR)
_logging.getLogger("aioxmpp").setLevel(_logging.ERROR)
_logging.getLogger("asyncio").setLevel(_logging.ERROR)



import importlib
import logging
import os
import sys
import time
from pathlib import Path
from typing import Type

from aioxmpp import JID
from spade import quit_spade

logger = logging.getLogger(__name__)


def boot_agent(
    file: Path,
    jid: JID,
    name: str,
    number: int,
    properties: Path,
    log_level: int,
    verify_security: bool,
):
    """Configures logging system and boots the agent.

    Args:
        file: File path where the agent's class is.
        jid: JID of the agent.
        name: The name of the agent.
        number: If the agent its a clone is the number of present in
            the name of the agent, else its None.
        properties: Properties to be injected in the agent.
        log_level: Logging level to be used in the agents logging file.
        verify_security: If true it validates the SSL certificates.
    """
    log_file_name: str = jid.localpart + ("_" + jid.resource if jid.resource else "")
    logs_folder = file.parent.absolute().joinpath("logs")
    log_file = logs_folder.joinpath(f"{log_file_name}.log")

    os.makedirs(logs_folder, exist_ok=True)
    logging.basicConfig(
        filename=log_file,
        filemode="w",
        level=log_level,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    _boot_agent(file, jid, name, number, properties, verify_security)


def _boot_agent(
    file: Path,
    jid: JID,
    name: str,
    number: int,
    properties: Path,
    verify_security: bool,
):
    """Boots the agent."""
    # Gets the agent and properties classes. Creates the agent with the
    # properties and the attributes already provided. Runs the agent and
    # creates a loop that waits until the agent dies.
    logger.debug("creating agent")
    agent_class = _get_class(file)
    file_abs_path = file.parent.absolute()
    if properties:
        os.chdir(properties.parent)
        properties = _get_class(properties)(jid.localpart, name, number)
        properties = properties.extract()
    else:
        properties = None

    os.chdir(file_abs_path)
    agent_instance = agent_class(jid, properties, verify_security)
    logger.info("starting agent")
    try:
        agent_instance.start().result()
    except Exception as err:
        logger.error("could not start the agent: {}".format(err))
    while agent_instance.is_alive():
        try:
            time.sleep(1)
        except Exception as e:
            logger.error("AGENT CRACHED")
            logger.exception(e)
            agent_instance.stop()
        except KeyboardInterrupt:
            logger.info("Keyboard Interrupt")
            agent_instance.stop()
    quit_spade()
    logger.info("agent stoped")


def _get_class(file: Path) -> Type:
    """Gets class from a file.

    Reads a python module and retrieves the class with the name of the file.


    Args:
        file: Python module. Must have a class with the same name as the file.
            Example: agent.py --> class agent(...)

    Returns:
        A class object with the same name as the file.
    """
    module_path, module_file = os.path.split(file.absolute())
    module_name = module_file.split(".")[0]
    sys.path.append(module_path)
    module = importlib.import_module(module_name)
    return getattr(module, module_name)



import logging
from time import sleep

from peak import DF


def exec(log_level: int, domain: str, verify_security: bool, port: int, *args, **kargs):
    """Executes the Directory Facilitator agent.

    Args:
        log_level: Logging level to be used in the DF's logs.
        domain: Domain to which connect the DF.
        verify_security: Verifies the SSL certificates.
        port: Port to be used by the DF REST API.
    """
    logging.basicConfig(
        level=log_level,
        format="%(asctime)s [%(levelname)s] %(message)s",
        handlers=[logging.StreamHandler()],
    )
    logger = logging.getLogger(__name__)
    logger.info("Starting DF")
    df = DF(domain, verify_security, port)
    df.start().result()
    logger.info("DF running")

    try:
        while df.is_alive():
            sleep(10)
    except KeyboardInterrupt:
        df.stop()
    logger.info("Stoped DF")



from abc import ABCMeta, abstractmethod
from random import uniform
from typing import Dict, Iterable

from pandas import DataFrame

from peak.drivers import Driver


class Property:
    """Represents a property of an agent.

    Attributes:
        data: Array of objects.
        loop: If True, when the property reaches the end of the data array
            it comes back to the begining.
        offset: Offset of the index of the data array.
        random_range: Percentage that applies randomness to the data.
        current_value: Current value of the property.
    """

    def __init__(
        self,
        data: Iterable,
        loop: bool = False,
        random_range: float = None,
        offset: int = 0,
    ):
        self.data = data
        self.loop = loop
        self.offset = offset
        self.random_range = random_range
        self.iter = self._gen(data, loop, offset)
        self.current_value = next(self.iter)

    def _gen(self, data: Iterable, loop: bool, offset: int):
        """Retrieves values from the data iteratively.

        Args:
            data: Array of objects.
            loop: If True, when the property reaches the end of the data array
                it comes back to the begining.
            offset: Offset of the index of the data array.

        Yields:
            The next value in the data array.
        """
        f = True
        while f:
            try:
                data = data[offset:]
                for i in data:
                    if self.random_range:
                        yield i * uniform(1 - self.random_range, 1 + self.random_range)
                    else:
                        yield i

                if not loop:
                    while True:
                        yield None
            except TypeError:
                f = False
        while True:
            yield data

    def next(self):
        """Iterates the data array."""
        self.current_value = next(self.iter)

    def __add__(self, other):
        if isinstance(other, Property):
            return self.current_value + other.current_value
        if isinstance(other, (float, int)):
            return self.current_value + other
        raise TypeError

    def __sub__(self, other):
        if isinstance(other, Property):
            return float(self.current_value) - float(other.current_value)
        if isinstance(other, (float, int)):
            return self.current_value - other
        raise TypeError

    def __mul__(self, other):
        if isinstance(other, Property):
            return self.current_value * other.current_value
        if isinstance(other, (float, int)):
            return self.current_value * other
        raise TypeError

    def __div__(self, other):
        if isinstance(other, Property):
            return self.current_value / other.current_value
        if isinstance(other, (float, int)):
            return self.current_value / other
        raise TypeError

    def __repr__(self) -> str:
        return str(self.current_value)

    def __getitem__(self, item):
        return self.data[(item % len(self.data)) - 1]


class Properties(metaclass=ABCMeta):
    """Used to filtrate and process the data to be injected in the agent.

    Attributes:
        agent_fullname: If the agent is a clone the full name will have
            its original name the the clone number. If its not, the full
            name will be equal to the name.
        agent_name: Just the name without the clone number.
        agent_number: The clone number if its a clone.
    """

    def __init__(
        self, full_name: str = None, name: str = None, number: int = None
    ) -> None:
        self.agent_fullname: str = full_name
        self.agent_name: str = name
        self.agent_number: int = number
        self.ds = dict()
        self.build_dataset()

    def add_property(self, property_name: str, property: Property):
        """Adds a single Property to the agent's properties.

        Args:
            property_name: Name of the property.
            property: An instance of Property.
        """
        if self.agent_fullname not in self.ds:
            self.ds[self.agent_fullname] = dict()
        self.ds[self.agent_fullname][property_name] = property

    def add_dataset(
        self,
        dataframe: DataFrame,
        loop: bool = False,
        random_range: float = None,
        offset: int = 0,
    ):
        """Adds a dataframe to the agent's properties.

        Args:
            dataframe: Pandas Dataframe.
            loop: If True, when the property reaches the end of the data array
                it comes back to the begining.
            random_range: Percentage that applies randomness to the data.
            offset: Offset of the index of the data array.
        """
        if self.agent_fullname not in self.ds:
            self.ds[self.agent_fullname] = dict()
        for property in dataframe:
            self.ds[self.agent_fullname][property] = Property(
                dataframe[property].to_list(), loop, random_range, offset
            )

    def add_driver(self, driver_name: str, driver: Driver):
        """Adds a driver to the agent properties.

        Args:
            driver_name: Name of the driver.
            driver: The actual driver.
        """
        if self.agent_fullname not in self.ds:
            self.ds[self.agent_fullname] = dict()
        self.ds[self.agent_fullname][driver_name] = driver

    @abstractmethod
    def build_dataset(self):
        """Contains the process of filtering and adding the data to the agent."""
        pass

    def extract(self) -> Dict[str, Property]:
        """Extracts the dataset from this properties.

        Returns:
            A dictionary with the agents properties.
        """
        return self.ds[self.agent_fullname]



from argparse import ArgumentError, ArgumentTypeError
from logging import getLevelName, getLogger
from multiprocessing import Process
from os import chdir
from pathlib import Path
import yaml

from peak import JID
from peak.bootloader import boot_agent


_logger = getLogger()

def agent_exec(
    file: Path,
    jid: JID,
    properties: Path = None,
    clones: int = 1,
    log_level: int = getLevelName("INFO"),
    verify_security: bool = False,
    *args,
    **kargs
):
    """Executes and configures a single agent.

    The arguments args and kargs are used to ignore extra parameters given be the parser.

    Args:
        file: Path to the agent's python file.
        properties: Path to the agent's properties python file.
        jid: JID of the agent.
        clones: Number of clones to be made.
        log_leve: Logging level.
        verify_security: Verifies the SSL certificates.

    Raises:
        ArgumentTypeError if the file provided is not a python file.
    """

    log_level = getLevelName(log_level)
    for f in [file, properties]:
        if f and not f.is_file():
            raise ArgumentTypeError("'{}' must be a python file".format(f))

    kwargs = {
        "file": file,
        "jid": jid,
        "name": jid.localpart,
        "number": None,
        "properties": properties,
        "log_level": log_level,
        "verify_security": verify_security,
    }

    if clones == 1:
        boot_agent(**kwargs)
    else:
        procs = []
        for i in range(clones):
            kwargs["jid"] = jid.replace(localpart=jid.localpart + str(i))
            kwargs["number"] = i
            proc = Process(target=boot_agent, kwargs=kwargs, daemon=False)
            proc.start()
            procs.append(proc)
        for proc in procs:
            try:
                proc.join()
            except Exception as e:
                _logger.exception(e)
            except KeyboardInterrupt:
                break


def multi_agent_exec(file: Path, *args, **kargs):
    """Executes multiple agents using a YAML configuration file.

    The arguments args and kargs are used to ignore extra parameters given be the parser.

    Args:
        file: Path to the agent's python file.
    """
    defaults = {
        "file": None,
        "domain": None,
        "resource": None,
        "ssl": False,
        "log_level": "info",
        "properties": None,
        "clones": 1,
    }
    _logger.debug("parsing yaml file")
    with file.open() as f:
        yml = yaml.full_load(f)
    chdir(file.parent)
    if "defaults" in yml:
        defaults = defaults | yml["defaults"]
    else:
        _logger.debug('no defaults in yaml file')
    if "agents" not in yml:
        raise Exception("agents argument required")
    _logger.debug("initialize processes")
    procs = []
    for agent, agent_args in yml["agents"].items():
        agent_args = defaults | agent_args

        if agent_args["file"] is None:
            raise Exception(f"{agent}: file argument required")
        if agent_args["domain"] is None:
            raise Exception(f"{agent}: domain argument required")
        agent_args["file"] = Path(agent_args["file"])
        if agent_args["properties"]:
            agent_args["properties"] = Path(agent_args["properties"])
        agent_args["jid"] = JID(agent, agent_args["domain"], agent_args["resource"])
        agent_args["log_level"] = agent_args["log_level"].upper()

        proc = Process(target=agent_exec, kwargs=agent_args, daemon=False)
        proc.start()
        procs.append(proc)
    _logger.debug("wait for processes to finish")
    for proc in procs:
        try:
            proc.join()
        except Exception as e:
            _logger.exception(e)
        except KeyboardInterrupt:
            break
    

